package ca.uol.aig.fts.drpipeline;

import ca.uol.aig.fts.io.NDFIO;
import ca.uol.aig.fts.fitting.CubicSplineInterpolation;
import ca.uol.aig.fts.phasecorrection.PhaseCorrection;
import ca.uol.aig.fftpack.RealDoubleFFT_Even;
/**
 * DRPipeline acts as a data reduction pipeline. It will intergrate 
 * Interferogram/Spectrum I/O, Interpolation, PhaseCorrection,
 * and FFT into a pipeline.
 *
 * @author Baoshe Zhang
 * @author Astronomial Instrument Group of UoL
 */
public class DRPipeline 
{
      /**
       * Constructor
       * @param in the path of the raw data file
       * @param out the path of the reduced data file
       * @param pcfSize_h the half of the size of the phase correction function
       * @param dsSize the half of the size of the doube-size interferogram
       * @param fittingDegree the polynomial degree of phase fitting 
       * @param weight_limit when the amplitude of a point over the amplitude
       *        maxima is greater than weigth_limit, this point will be taken
       *        into account in phase fitting
       */
      public DRPipeline(String in, String out, int pcfSize_h, 
                        int dsSize, int fittingDegree, double weight_limit)
      {
            /* get the raw data from an interferogram file,
               create a new spectrum file
            */
            long t0 = System.currentTimeMillis();

            NDFIO ndf_ifgm = new NDFIO(in, out);

            /* get the original irregular mirror positions */
            double[] mirrorPos = ndf_ifgm.getMirrorPos();

            /* create the new regular mirror positions */
            double[] new_mirrorPos = new double[mirrorPos.length];
            double x_interval = (mirrorPos[mirrorPos.length-1] - mirrorPos[0])/(mirrorPos.length-1);
            new_mirrorPos[0] = mirrorPos[0];
            for(int i=1; i<mirrorPos.length-1; i++)
                new_mirrorPos[i] = new_mirrorPos[i-1] + x_interval;
            new_mirrorPos[new_mirrorPos.length-1] = mirrorPos[mirrorPos.length-1];

            long t1 = System.currentTimeMillis();
            System.out.println("Read Time : " + (t1 - t0));
            
            /* interpolation of interferograms */
            CubicSplineInterpolation csi2fts = new CubicSplineInterpolation(mirrorPos, new_mirrorPos);
            double[] single_ifgm;
            int arrayWidth = ndf_ifgm.get_arrayWidth();
            int arrayLength = ndf_ifgm.get_arrayLength();
            double[][][] ifgm_interp = new double[arrayWidth][arrayLength][];

            for(int i=0; i<arrayWidth; i++)
               for(int j=0; j<arrayLength; j++)
               {
                     single_ifgm = ndf_ifgm.getInterferogram(i, j);
                     ifgm_interp[i][j] = csi2fts.interpolate(single_ifgm);
               } 

            long t2 = System.currentTimeMillis();
            System.out.println("Interpolation Time : " + (t2 - t1));

            /* phase correction of interferograms */

            int ssSize = ifgm_interp[0][0].length - dsSize + 1;
            PhaseCorrection pc2fts = new PhaseCorrection(dsSize, ssSize, fittingDegree, 
                                                     pcfSize, weight_limit); 

            double[][][] ifgm_pc = new double[arrayWidth][arrayLength][];
            for(int i=0; i<arrayWidth; i++)
               for(int j=0; j<arrayLength; j++)
               {
                     ifgm_pc[i][j] = pc2fts.getInterferogram(ifgm_interp[i][j]);
               }

            long t3 = System.currentTimeMillis();
            System.out.println("Phase Corretion Time : " + (t3 - t2));

            /* FFT of interferograms */
            RealDoubleFFT_Even fft2fts = new RealDoubleFFT_Even(ifgm_pc[0][0].length);
            for(int i=0; i<arrayWidth; i++)
               for(int j=0; j<arrayLength; j++)
               {
                     fft2fts.ft(ifgm_pc[i][j]);
               }

            long t4 = System.currentTimeMillis();
            System.out.println("FFT Time : " + (t4 - t3));

            /* save the spectrum to a file */

            int arraySize = arrayWidth * arrayLength;
            int nPoints_Spectrum = ifgm_pc[0][0].length;
            int cubeSize = arraySize * nPoints_Spectrum; 
            float[] cube_spectrum = new float[cubeSize];
            int index = 0;
            for(int k = 0; k < nPoints_Spectrum; k++)
               for(int j=0; j<arrayLength; j++)
                  for(int i=0; i<arrayWidth; i++)
                  {
                        cube_spectrum[index] = (float)ifgm_pc[i][j][k];
                        index++;
                  }
            long[] spectrum_cube_shape = new long[3];
            spectrum_cube_shape[0] = arrayWidth;
            spectrum_cube_shape[1] = arrayLength;
            spectrum_cube_shape[2] = nPoints_Spectrum;
            ndf_ifgm.saveSpectrum(cube_spectrum, spectrum_cube_shape);

            ndf_ifgm.closeSpectrum();

            long t5 = System.currentTimeMillis();
            System.out.println("Write Time : " + (t5 - t4));
            System.out.println("Total Time : " + (t5 - t0));
      }
}
