package ca.uol.aig.fts.fitting;
/**
 *  Optimized CubicSpline Interpolator
 *  @author Baoshe Zhang
 *  @author Astronomical Instrument Group of UoL
 */
public class CubicSplineInterpolation 
{
    private double[] _coeff, _s, _idx, _p, _coeff_x, _idx26, _dxdx6;
    private double[] _x, _y;
    private int _n;
    private double[] _x_new;
    private int[] _x_index;

    /**
     * Constructor.
     * @param x_orig the values of x-axis to be interpolate
     * @param x_new  the values of the interpolated x-axis
     */
    public CubicSplineInterpolation(double[] x_orig, double[] x_new) 
    {
          if((x_new[0] < x_orig[0]) || (x_new[x_new.length-1] > x_orig[x_orig.length - 1]))
          {
              throw new IllegalArgumentException("Abscissa out of range");
          }

          /* the following arrays or variables only related to x_orig */
          _n = x_orig.length;
          _x = x_orig;
          _s = new double[_n];
          _idx = new double[_n];
          _idx26 = new double[_n];
          _p  = new double[_n];
          _dxdx6 = new double[_n];
          _coeff_x = new double[_n];
          _x_new = x_new;

          
          _coeff_x[0] = 0.0;
          for(int i = 1; i < _n-1; i++) 
          {
             _s[i] = (_x[i] - _x[i-1]) / (_x[i+1] - _x[i-1]);
             _idx[i] = 1.0/(_x[i] - _x[i-1]);
             _idx26[i] = 6.0/(_x[i+1] - _x[i-1]);
             _p[i] = 1.0/(_s[i] * _coeff_x[i-1] + 2.0);
             _coeff_x[i] = (_s[i] - 1.0) * _p[i];
             _dxdx6[i] = (_x[i] - _x[i-1])*(_x[i] - _x[i-1])/6.0;
          }
          _idx[_n-1] = 1.0/(_x[_n-1] - _x[_n-2]);
          _dxdx6[_n-1] = (_x[_n-1] - _x[_n-2])*(_x[_n-1] - _x[_n-2])/6.0;


          calc_index();
    }    

    /* calculate the index of the new x-axis values. */
    private void calc_index()
    {
          /** Binary search for correct place in the table. */

          _x_index = new int[_x_new.length];

          int j, k;
          for(int index=0; index<_x_new.length; index++)
          {
              j = 0;
              k = _n-1;

              while (k - j > 1)
              {
                  int i = (k + j) >> 1;
                  if (_x[i] > _x_new[index]) k = i;
                  else j = i;
              }
              _x_index[index] = k;
          }
    }

    /* Use y_orig to initialize the coefficient array. */
    private void initialize(double[] y) 
    {
	  /* Solve for coefficients by tridiagonal algorithm. */

          _y = y;

	  double[] z = new double[_n];
	  z[0] = 0.0;
	
	  _coeff = new double[_n];

          double[] _dy = new double[_n];
          for(int i = 1; i < _n; i++) _dy[i] = _y[i] - _y[i-1];

          double t;
	  for(int i = 1; i < _n-1; i++) 
          {
	       t = _dy[i+1] * _idx[i+1] - _dy[i] * _idx[i];
	       z[i] = (t * _idx26[i] - _s[i]*z[i-1]) * _p[i];
	  }
	
	  _coeff[_n - 1] = 0.0;
	  for (int k = _n-2; k >= 0; k--) 
          {
	       _coeff[k] = _coeff_x[k] * _coeff[k+1] + z[k];
	  }
    }

    /* Interpolate a specified point indexed by i. */
    private double interpolate_(int i)
    {
          int k = _x_index[i];
          double a = (_x[k] - _x_new[i]) * _idx[k];
          double b = (_x_new[i] - _x[k-1]) * _idx[k];
          return a * _y[k-1] + b * _y[k]
                 + ((a*a*a - a) * _coeff[k-1] + (b*b*b - b) * _coeff[k])
                 * _dxdx6[k];
    }

    /**
     * Interpolate 
     * @param y_orig the values of y-axis corresponding to x_orig
     * @return the values of the interpolate y values corresponding to x_new
     */

    public double[] interpolate(double[] y_orig)
    {
           initialize(y_orig);

           double[] y_new = new double[_x_new.length];
           for(int i=0; i<_x_new.length; i++) y_new[i] = interpolate_(i);

           return y_new;
    }
}
