package ca.uol.aig.fts.phasecorrection;

import ca.uol.aig.fftpack.Complex1D;
import ca.uol.aig.fftpack.RealDoubleFFT;
import ca.uol.aig.fftpack.RealDoubleFFT_Even;
import ca.uol.aig.fftpack.RealDoubleFFT_Odd;
import ca.uol.aig.fts.fitting.PolynomialFitting;

/**
 * Correct the phases of single-interferograms
 * @author Baoshe Zhang
 * @author Astronomial Instrument Group of UoL
 */
public class PhaseCorrection
{
     int dsLength, ssLength, phaseFittingdegree, pcfSize;
     double weight_limit;
     RealDoubleFFT      pc_rfft   = null;
     RealDoubleFFT_Even pc_rfft_e = null;
     RealDoubleFFT_Odd  pc_rfft_o = null;

     /**
      * Constructor
      * @param dsLength the half of the length of double-side interferogram
      * @param ssLength the length of single-side interferogram
      * @param pcfSize_h the half of the size of the phase-correction function
      * @param weight_limit used to remove insignificant points. The points whose
      *      energy (the square of amplitude) is less than (weight_limit x maxima
      *      of energy) will be not taken into account in the phase-fitting.
      */
     public PhaseCorrection(int dsLength, int ssLength, int phaseFittingdegree, int pcfSize_h, 
                            double weight_limit)
     {
          this.dsLength = dsLength;
          this.ssLength = ssLength;
          this.phaseFittingdegree = phaseFittingdegree;
          this.pcfSize = pcfSize_h;
          this.weight_limit = weight_limit;

          pc_rfft   = new RealDoubleFFT(2*dsLength);
          pc_rfft_e = new RealDoubleFFT_Even(dsLength+1);
          pc_rfft_o = new RealDoubleFFT_Odd(dsLength-1);
     }

     /* smooth the phases and get rid of phase jitters */ 
     void phaseSmoothing(double[] phase)
     {
          int[] phase_jitter = new int[phase.length];

          phase_jitter[0] = 0;
          for(int i=1; i<phase.length; i++)
          {
               phase_jitter[i] = phase_jitter[i-1];
               if(Math.abs(phase[i]-phase[i-1]) > Math.PI*2.0/3.0)
               {
                   if(phase[i] > phase[i-1])phase_jitter[i]-=1;
                   else phase_jitter[i]+=1;
               }
          }
          for(int i=0; i<phase.length; i++)
             phase[i] = phase[i] + Math.PI * phase_jitter[i];
     }

     /* get the phases from double-side interferograms */
     double[] getPhase(double[] dsInterferogram)
     {

          double[] m_dsInterferogram = new double[2*dsLength];

          boolean new_PCF;
          new_PCF = true;
          if(new_PCF)
          {
              System.arraycopy(dsInterferogram, 0, m_dsInterferogram, 0, 2*dsLength);
          }
          else
          {
              System.arraycopy(dsInterferogram, dsLength-1, m_dsInterferogram, 0, dsLength+1);
              System.arraycopy(dsInterferogram, 0, m_dsInterferogram, dsLength+1, dsLength-1);
          }

          Complex1D cSpectrum = new Complex1D();

          pc_rfft.ft(m_dsInterferogram, cSpectrum);

          double[] weights = new double[cSpectrum.x.length];

          weights[0] = 0.0D;
          double weight_max = 0;
          for(int i=1; i<cSpectrum.y.length-1; i++)
          {
               weights[i] = cSpectrum.x[i] * cSpectrum.x[i] + cSpectrum.y[i] * cSpectrum.y[i];
               if(weights[i] > weight_max) weight_max = weights[i];
          }
          weights[cSpectrum.x.length-1] = 0.0;

          double[] phase = new double[cSpectrum.x.length];
          double phase_t;
          for(int i=0; i<cSpectrum.x.length; i++)
          {
               phase_t = Math.atan2(cSpectrum.y[i], cSpectrum.x[i]);
               if(new_PCF)
               {
                    phase_t -= Math.PI/(cSpectrum.x.length-1) * i;
               }

               if(phase_t < 0.000)
               {
                    phase_t += Math.PI;
                    if(phase_t < 0.00) phase_t += Math.PI;
               }
               phase[i] = phase_t;
          }
          phaseSmoothing(phase);

          double weights_abs_min = weight_max * weight_limit;

          double[] phase_weights = new double[3*cSpectrum.x.length+1];
          int coord_index = 0;
          int phase_index = cSpectrum.x.length;
          int weight_index = 2*cSpectrum.x.length;

          for(int i=0; i<cSpectrum.x.length; i++)
          {
               if(weights[i] > weights_abs_min)
               { 
                   phase_weights[coord_index++] = i;
                   phase_weights[phase_index++] = phase[i];
                   phase_weights[weight_index++] = weights[i];
               }
          }
          phase_weights[3*cSpectrum.x.length] = coord_index;

          return phaseFitting(phase_weights);
     }

     /* get the new phases from weighted fitting */
     double[] phaseFitting(double[] phase_weights)
     {
          int NPoints = (int)phase_weights[phase_weights.length-1];
          double[] wavenumber = new double[NPoints];
          double[] phase = new double[NPoints];
          double[] weights = new double[NPoints];
          for(int i=0; i<NPoints;i++)
          {
               wavenumber[i] = phase_weights[i];
               phase[i] = phase_weights[dsLength+1+i];
               weights[i] = phase_weights[2*(dsLength+1)+i];
//               System.out.println(wavenumber[i] + " " + phase[i]);
          }
          PolynomialFitting pnf = new PolynomialFitting(phaseFittingdegree);
          pnf.fit(wavenumber, phase, weights);

          double[] new_wavenumber = new double[dsLength+1];
          for(int i=0; i<dsLength+1; i++) new_wavenumber[i] = i;

          double[] new_phase = pnf.getResult(new_wavenumber);
/*
          for(int i=0; i<new_phase.length; i++)
             System.out.println(i + " " + new_phase[i]);
          System.exit(0);
*/

          return new_phase;
     }

     /* get a new phase-correction function from the phases */
     double[] calcPCF(double[] phase)
     {
          double[] cos_phase = new double[dsLength+1];
          double[] sin_phase = new double[dsLength-1];

          for(int i=0; i<dsLength+1; i++)
          {
               cos_phase[i] = Math.cos(phase[i]);
          }

          for(int i=0; i<dsLength-1; i++)
          {
               sin_phase[i] = Math.sin(phase[i+1]);
          }

          pc_rfft_e.bt(cos_phase);
          pc_rfft_o.bt(sin_phase);

          double[] pcf;
          if(pcfSize >= dsLength)
          {
              pcf = new double[2*dsLength];

              pcf[dsLength-1] = cos_phase[0];
              for(int i=1; i<dsLength; i++)
              {
                 pcf[dsLength-1+i] = cos_phase[i] + sin_phase[i-1];
                 pcf[dsLength-1-i] = cos_phase[i] - sin_phase[i-1];
              }
              pcf[2*dsLength-1] = cos_phase[dsLength];
          }
          else
          {
              pcf = new double[2*pcfSize];

              pcf[pcfSize-1] = cos_phase[0];
              for(int i=1; i<pcfSize; i++)
              {
                 pcf[pcfSize-1+i] = cos_phase[i] + sin_phase[i-1];
                 pcf[pcfSize-1-i] = cos_phase[i] - sin_phase[i-1];
              }
              pcf[2*pcfSize-1] = cos_phase[pcfSize] + sin_phase[pcfSize-1];
          }

          return pcf;
     }

     /* convolution between x and y. */
     double[] conv(double[] x, double[] y, int offset, int ext_right)
     {

         double temp;
         int k;

         if(x.length < y.length)
             throw new IllegalArgumentException("The length of x should not be less than that of y!");
         if(offset<0 || offset > x.length)
             throw new IllegalArgumentException("The value of offset is invalid!");

         double[] value = new double[x.length+y.length/2-1+ext_right-offset];
         for(int i=offset; i<x.length+y.length/2-1+ext_right; i++)
         {
              temp = 0;
              for(int j=0; j<y.length; j++)
              {
                   k = i - j;
                   if(k < 0) k += x.length;
//                   if(k > x.length) k -= x.length;
                   if(k < x.length) temp = temp + x[k]*y[j];
              }
              value[i-offset] = temp;
         }
         return value;
     }

     /**
      * get the phase-corrected one-side interferogram
      * @param compositeInterferogram the composite interferogram, including a short double-side
      *                   and a long one-side
      * @return the phose-corrected one-side interferogram with ext_right right extensions. The
      *    length of the phase-corrected interferogram is equal to (ssLength + ext_right) 
      */
     public double[] getInterferogram(double[] compositeInterferogram)
     {
          double[] dsInterferogram = new double[2*dsLength];
          System.arraycopy(compositeInterferogram, 0, dsInterferogram, 0, 2*dsLength);
          double[] ssInterferogram = new double[dsLength+ssLength-1];
          System.arraycopy(compositeInterferogram, 0, ssInterferogram, 0, dsLength+ssLength-1);

          double[] phase = getPhase(dsInterferogram);

          double[] pcf = calcPCF(phase);

          return conv(ssInterferogram, pcf, dsLength+pcfSize-2, 0);
     }
}
